##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
__doc__ = """Utility for locating memory leaks.  Gorilla patch galore!"""
__version__ = '$Revision: 1.3 $'[11:-2]

import sys
from string import join
from thread import allocate_lock
from time import ctime, time, sleep
from traceback import format_stack

import Acquisition
from Acquisition import aq_inner, aq_parent
import Globals
from ZODB.DB import DB
from ZODB.cPickleCache import PickleCache
from OFS.SimpleItem import SimpleItem
from Globals import DTMLFile


if not hasattr(DB, '_lf_real_open'):
    # Patch DB with a way to block open operations.
    DB._open_lock = allocate_lock()
    def open(self, *args, **kw):
        self._open_lock.acquire()
        self._open_lock.release()
        return apply(self._lf_real_open, args, kw)
    DB._lf_real_open = DB.open
    DB.open = open
        

class dummyclass: pass
classtypes = (type(dummyclass), type(Acquisition.Implicit))


def refcounts():
    '''
    Returns a list containing (refcount, classname, classob)
    '''
    t = classtypes
    dict={}
    for m in sys.modules.values():
        for sym in dir(m):
            ob=getattr(m, sym)
            if type(ob) in t:
                dict[ob]=sys.getrefcount(ob)
    pairs=[]
    append=pairs.append
    for ob, v in dict.items():
        if hasattr(ob, '__module__'):
            name='%s.%s' % (ob.__module__, ob.__name__)
        else: name='%s' % ob.__name__
        append((v, name, ob))
    return pairs

def getControlledRefcounts(db, my_conn=None, my_app=None, timeout=5, gc=0):
    '''
    Momentarily blocks the creation of new connections, then zaps
    connection caches so that a "controlled" sample can be achieved.
    Returns (refcounts, allocated_connection_count, problem_connection_count)
    '''
    db._open_lock.acquire()
    expected = -1  # Expect one connection to be open.
    try:
        pools, pooll = db._pools
        current = 0
        for version, (pool, allocated, pool_lock) in pools.items():
            expected = expected + len(allocated)
            current = current + len(pool)
        expiration = time() + timeout
        while current < expected and time() < expiration:
            sleep(1)
            current = 0
            for version, (pool, allocated, pool_lock) in pools.items():
                current = current + len(pool)
        # Count the number of connections that didn't close.
        problem_connection_count = expected - current
        # Clear the cache in each closed connection.
        for version, (pool, allocated, pool_lock) in pools.items():
            for c in pool:
                resetCache(c, gc)
        if my_app is not None:
            # Deactivate the application object.
            if hasattr(my_app, 'acl_users') and hasattr(my_app.acl_users,
                                                        '_p_deactivate'):
                my_app.acl_users._p_deactivate()
            my_app.Control_Panel._p_deactivate()
            my_app._p_deactivate()
        if my_conn is not None:
            # Clear cache in own connection (meaning that if
            # any writes happen to occur during this transaction, ZODB
            # could do something wrong.  Oh well.)
            resetCache(my_conn, gc)
        # We're now in a controlled environment.
        counts = refcounts()
    finally:
        # Restore normal operations.
        db._open_lock.release()
    return counts, expected + 1, problem_connection_count

def resetCache(c, gc):
    cache = c._cache
    if gc:
        cache.minimize(3)  # The minimum that actually performs gc.
    # Replace it with an empty cache.
    c._cache = PickleCache(c, cache.cache_size, cache.cache_age)


patched = {}  # {name -> class}
constructions = {}  # {id -> string}
ref_sample = None  # {name -> count}

def patched_init(self, *args, **kw):
    tb = join(format_stack(), '')
    now = ctime(time())
    klass = self.__class__
    sid = id(self)
    constructions[sid] = '%s.%s - %s (%s):\n%s' % (
        getattr(klass, '__module__', '?'), getattr(klass, '__name__', '?'),
        now, sid, tb)
    d = klass.__dict__
    if d.has_key('__lf_inh_init__'):
        return apply(d['__lf_inh_init__'], (self,) + args, kw)

def patched_del(self, constructions=constructions, *args, **kw):
    try: del constructions[id(self)]
    except KeyError: pass
    d = self.__class__.__dict__
    if d.has_key('__lf_inh_del__'):
        return apply(d['__lf_inh_del__'], (self,) + args, kw)
    

class LeakFinder (SimpleItem):
    '''
    '''
    meta_type = id = title = 'LeakFinder'

    manage_options = (
        {'label':'Usage', 'action':'manage_usage'},
        {'label':'RefCounts', 'action':'manage_refcounts'},
        {'label':'Results', 'action':'manage_results'},
        )

    manage_usage = DTMLFile('usage', globals())
        
    manage_refcounts = DTMLFile('refcounts', globals())

    manage_results = DTMLFile('results', globals())

    def manage_getSample(self, timeout=5, gc=0, reset_reference=0,
                         show_unchanged=0):
        '''
        '''
        global ref_sample
        app = self.getPhysicalRoot()
        conn = app._p_jar
        db = conn._db
        counts, allocated_c, problem_c = getControlledRefcounts(
            db, conn, app, timeout=int(timeout), gc=gc)
        diffs = []
        if ref_sample is None:
            reset_reference = 1
        for count, name, ob in counts:
            if count > 3:
                if reset_reference:
                    diff = 0
                else:
                    diff = count - ref_sample.get(name, 0)
                    if diff == 0 and not show_unchanged:
                        continue
                diffs.append((diff, count, name))
        diffs.sort()
        diffs.reverse()
        if reset_reference:
            ref_sample = {}
            for count, name, ob in counts:
                ref_sample[name] = count
        return {'diffs': diffs,
                'allocated_connections': allocated_c,
                'problem_connections': problem_c,
                'capture_classes': patched.keys(),
                }

    def manage_setCaptureClasses(self, capture_classes=(),
                                 remove_all_patches=0, REQUEST=None):
        '''
        '''
        global patched
        new_patched = {}
        for count, name, ob in refcounts():
            if not remove_all_patches and name in capture_classes:
                new_patched[name] = ob
        # Uninstall patches.
        for klass in patched.values():
            d = klass.__dict__
            if d.has_key('__lf_real_del__'):
                d['__del__'] = d['__lf_real_del__']
                del d['__lf_real_del__']
            elif d.has_key('__del__'):
                del d['__del__']
            if d.has_key('__lf_inh_del__'):
                del d['__lf_inh_del__']

            if d.has_key('__lf_real_init__'):
                d['__init__'] = d['__lf_real_init__']
                del d['__lf_real_init__']
            elif d.has_key('__init__'):
                del d['__init__']
            if d.has_key('__lf_inh_init__'):
                del d['__lf_inh_init__']

        # Install patches.
        patched = new_patched
        for name, klass in new_patched.items():
            d = klass.__dict__
            d['__classname__'] = name
            if d.has_key('__del__'):
                d['__lf_real_del__'] = d['__del__']
                d['__lf_inh_del__'] = d['__del__']
            elif hasattr(klass, '__del__'):
                d['__lf_inh_del__'] = klass.__del__
            d['__del__'] = patched_del
            if d.has_key('__init__'):
                d['__lf_real_init__'] = d['__init__']
                d['__lf_inh_init__'] = d['__init__']
            elif hasattr(klass, '__init__'):
                d['__lf_inh_init__'] = klass.__init__
            d['__init__'] = patched_init
        if REQUEST is not None:
            return Globals.MessageDialog(REQUEST,
                                         title='LeakFinder',
                                         message='Patched and unpatched '
                                         'classes.',
                                         action='manage_refcounts')

    def manage_getCaptureResults(self):
        res = constructions.values()
        res.sort()
        return res

    def manage_clearResults(self):
        constructions.clear()


Globals.InitializeClass(LeakFinder)
